from aws_cdk import Stack, CustomResource, RemovalPolicy

from b_stage_deployment.function import StageDeploymentSingletonFunction


class StageDeploymentResource(CustomResource):
    """
    Custom resource used for create a custom stage deployment.
    """

    def __init__(
            self,
            scope: Stack,
            resource_name: str,
            deployment_function: StageDeploymentSingletonFunction,
            api_id: str,
            stage_name: str,
            description: str
    ) -> None:
        """
        Constructor.

        :param scope: CloudFormation template stack in which this resource will belong.
        :param resource_name: Custom name for a resource.
        :param deployment_function: Resource function.
        :param api_id: Identification string of an api for which a deployment should be managed.
        :param stage_name: The name of the stage belonging to an api.
        :param description: Deployment description.
        """
        super().__init__(
            scope=scope,
            id=resource_name,
            service_token=deployment_function.function_arn,
            pascal_case_properties=True,
            removal_policy=RemovalPolicy.DESTROY,
            properties={
                'ApiId': api_id,
                'StageName': stage_name,
                'Description': description
            }
        )



from functools import lru_cache

from aws_cdk import Stack, Duration
from aws_cdk.aws_iam import PolicyStatement
from aws_cdk.aws_lambda import Code, SingletonFunction, Runtime


class StageDeploymentSingletonFunction(SingletonFunction):
    """
    Custom api gateway stage deployment resource singleton lambda function.
    """

    def __init__(
            self,
            scope: Stack,
            name: str
    ) -> None:
        """
        Constructor.

        :param scope: CloudFormation stack in which this function will be deployed.
        :param name: The name of the function.
        """
        self.__name = name

        super().__init__(
            scope=scope,
            id=name,
            uuid=f'{name}-uuid',
            function_name=name,
            code=self.__code(),
            handler='index.handler',
            runtime=Runtime.PYTHON_3_8,
            timeout=Duration.minutes(1),
        )

        # Add permission to create deployments. Since this is a singleton lambda function,
        # we can not specify a specific api gateway resource.
        self.add_to_role_policy(PolicyStatement(
            actions=['apigateway:POST', 'apigateway:PATCH'],
            resources=['*']
        ))

    @lru_cache
    def __code(self) -> Code:
        """
        Gets (and caches) source code cor the lambda function.

        :return: Lambda function source code (as an asset).
        """
        from .source import root
        return Code.from_asset(root)

    @property
    def function_name(self) -> str:
        """
        Overrides original function_name function so the method would not create a dependency.

        :return: Name of the function.
        """
        return self.__name



import json
import logging
import boto3
from typing import Dict, Any, Optional, Tuple

logger = logging.getLogger(__name__)


class Action:
    def __init__(self, invocation_event: Dict[str, Any]):
        self.__invocation_event: Dict[str, Any] = invocation_event
        self.__parameters: Dict[str, Any] = invocation_event['ResourceProperties']
        self.__api_id = self.__parameters['ApiId']
        self.__stage_name = self.__parameters['StageName']
        self.__description = self.__parameters['Description']
        self.__client = boto3.client('apigatewayv2')

    def create(self) -> Tuple[Optional[Dict[Any, Any]], Optional[str]]:
        """
        Creates a resource.

        :return: A tuple containing two items:
            1. Custom data to return back to CloudFormation service.
            2. Physical resource id (can be empty).
        """
        logger.info(f'Initiating resource creation with these parameters: {json.dumps(self.__parameters)}.')

        self.__client.create_deployment(
            ApiId=self.__api_id,
            Description=self.__description,
            StageName=self.__stage_name
        )

        return {}, None

    def update(self) -> Tuple[Optional[Dict[Any, Any]], Optional[str]]:
        """
        Updates a resource.

        :return: A tuple containing two items:
            1. Custom data to return back to CloudFormation service.
            2. Physical resource id (can be empty).
        """
        logger.info(f'Initiating resource update with these parameters: {json.dumps(self.__parameters)}.')

        # Creation and update should both result in a new deployment.
        return self.create()

    def delete(self) -> Tuple[Optional[Dict[Any, Any]], Optional[str]]:
        """
        Actually it makes no sense to delete a deployment. What does it even mean to delete
        a deployment? In a matter of fact, if you attempt to delete a deployment without
        deleting a stage first, you would get an error from CloudFormation:

        Active stages pointing to this deployment must be moved or deleted
        (Service: AmazonApiGatewayV2; Status Code: 400; Error Code: BadRequestException;
        Request ID: <some-random-generated-id>; Proxy: null)

        Hence, this function does nothing, effectively solving all bugs that original
        deployment resource has.

        :return: A tuple containing two items:
            1. Custom data to return back to CloudFormation service.
            2. Physical resource id (can be empty).
        """
        return {}, None



import json
import logging
from typing import Dict, Any, Tuple, Optional

import boto3
import botocore
from botocore.exceptions import ClientError

logger = logging.getLogger()
logger.setLevel(logging.INFO)

try:
    from cfnresponse import CfnResponse
    from action import Action
except ImportError as ex:
    logger.exception('Failed import.')
    from b_stage_deployment.source.action import Action
    from b_stage_deployment.source.cfnresponse import CfnResponse

logger.info(f'Version of boto3 lib: {boto3.__version__}.')
logger.info(f'Version of botocore lib: {botocore.__version__}.')


def __handle(event, context) -> Tuple[Optional[Dict[Any, Any]], Optional[str]]:
    """
    Handles incoming event by invoking a specific action according to a request type.

    :param event: Invocation event.
    :param context: Invocation context.

    :return: A tuple containing two items:
        1. Custom data to return back to CloudFormation service (can be empty).
        2. Physical resource id (can be empty).
    """
    serialized_event = json.dumps(event, default=lambda o: '<not serializable>')
    logger.info(f'Got new request. Event: {serialized_event}.')

    action = Action(event)

    if event['RequestType'] == 'Create':
        return action.create()

    if event['RequestType'] == 'Update':
        return action.update()

    if event['RequestType'] == 'Delete':
        return action.delete()

    raise KeyError('Unsupported request type! Type: {}'.format(event['RequestType']))


def handler(event, context) -> None:
    """
    Handles incoming event.

    :param event: Invocation event.
    :param context: Invocation context.

    :return: No return.
    """
    response = CfnResponse(event, context)

    try:
        data, resource_id = __handle(event, context)
        response.respond(CfnResponse.CfnResponseStatus.SUCCESS, data=data, resource_id=resource_id)
    except ClientError as ex:
        err_msg = f'{repr(ex)}:{ex.response}'
        response.respond(CfnResponse.CfnResponseStatus.FAILED, status_reason=err_msg)
    except Exception as ex:
        response.respond(CfnResponse.CfnResponseStatus.FAILED, status_reason=repr(ex))



import logging
import json
from enum import Enum
from typing import Dict, Any, Optional


class CfnResponse:
    """
    Class that sends response back to the CloudFormation service. Read more about responses here:
    https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/crpg-ref-responses.html
    """

    class CfnResponseStatus(Enum):
        """
        Enum class specifying possible states for a response status.
        """
        SUCCESS = 'SUCCESS'
        FAILED = 'FAILED'

    def __init__(self, invocation_event: Dict[str, Any], context: Any):
        """
        Constructor.

        :param invocation_event: Event dictionary that was passed to this lambda function.
        :param context: Lambda context that was passed to this lambda function.
        """
        self.__invocation_event = invocation_event
        self.__context = context

    def respond(
            self,
            status: CfnResponseStatus,
            status_reason: Optional[str] = None,
            data: Optional[Dict[str, Any]] = None,
            resource_id: Optional[str] = None,
    ) -> None:
        """
        Creates and sends response back to CloudFormation service.

        :param status: Operation status - failed or success.
        :param status_reason: Reason message for the status.
        :param data: Dictionary data to return back to CloudFormation.
        :param resource_id: Specify a custom id for a resource(s). This value should be an identifier unique to
        the custom resource vendor, and can be up to 1 Kb in size. The value must be a non-empty string and must
        be identical for all responses for the same resource.

        :return: No return.
        """
        response_body = dict(
            Status=status.value,
            Reason=status_reason or f'See the details in CloudWatch: {self.__context.log_stream_name}.',
            PhysicalResourceId=resource_id or self.__invocation_event.get('PhysicalResourceId') or self.__context.log_stream_name,
            StackId=self.__invocation_event['StackId'],
            RequestId=self.__invocation_event['RequestId'],
            LogicalResourceId=self.__invocation_event['LogicalResourceId'],
            NoEcho=False,
            Data=data or {}
        )

        response_json = json.dumps(response_body, default=lambda o: '<not serializable>')
        response_url = self.__invocation_event['ResponseURL']
        self.__send(response_url, response_json)

    @staticmethod
    def __send(url: str, message: str) -> None:
        """
        Sends a json message to a specified url.

        :param url: A destination for the message.
        :param message: Message payload.

        :return: No return.
        """
        import urllib3

        http = urllib3.PoolManager()

        try:
            logging.info(f'Callback data: {message}.')
            r = http.request('PUT', url, body=message, headers={'Content-Type': 'application/json'})
            logging.info(f'Status code: {str(r.data)}.')
        except Exception as e:
            logging.exception(f'Callback PUT failed: {repr(e)}.')


import logging

from b_aws_testing_framework.credentials import Credentials
from b_stage_deployment_test.testing_infrastructure import TestingInfrastructure

logger = logging.getLogger(__name__)


def test_deployed() -> None:
    """
    Tests that deployments were created and stage was deployed.

    :return: No return.
    """
    response = Credentials().boto_session.client('apigatewayv2').get_deployments(
        ApiId=TestingInfrastructure.get_output('ApiId'),
        MaxResults='25'
    )

    items = response['Items']

    print(len(items))
    print(items)

    assert len(list(items)) >= 5

    for deployment in items:
        assert deployment['AutoDeployed'] is False
        assert deployment['DeploymentStatus'] == 'DEPLOYED'
